/*
* SACD Decoder plugin
* Copyright (c) 2011 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _SACD_DSDIFF_H_INCLUDED
#define _SACD_DSDIFF_H_INCLUDED

#include <stdint.h>
#include <foobar2000.h>
#include "endianess.h"
#include "scarletbook.h"
#include "sacd_reader.h"

#define MARK_TIME(m) ((double)m.hours * 60 * 60 + (double)m.minutes * 60 + (double)m.seconds + ((double)m.samples + (double)m.offset) / (double)m_samplerate)

using namespace pfc;

#pragma pack(1)

class ID {
public:
	uint8_t ckID[4];
	bool has_id(const char* id) {
		return ckID[0] == id[0] && ckID[1] == id[1] && ckID[2] == id[2] && ckID[3] == id[3];
	}
	void set_id(const char* id) {
		ckID[0] = id[0];
		ckID[1] = id[1];
		ckID[2] = id[2];
		ckID[3] = id[3];
	}
};

class Chunk : public ID {
public:
	uint64_t ckDataSize;
	uint64_t get_size() {
		return hton64(ckDataSize);
	}
	void set_size(uint64_t size) {
		ckDataSize = hton64(size);
	}
};

class FormDSDChunk : public Chunk {
public:
	ID formType;
};

class DSTFrameIndex {
public:
	uint64_t offset;
	uint32_t length;
};

enum MarkType {TrackStart = 0, TrackStop = 1, ProgramStart = 2, Index = 4};

class Marker {
public:
	uint16_t hours;
	uint8_t  minutes;
	uint8_t  seconds;
	uint32_t samples;
	int32_t  offset;
	uint16_t markType;
	uint16_t markChannel;
	uint16_t TrackFlags;
	uint32_t count;
};

class subsong_t {
public:
	double start_time;
	double stop_time;
};

class id3tags_t {
public:
	uint32_t subsong;
	uint64_t offset;
	uint64_t size;
	array_t<uint8_t> data;
};

#pragma pack()

class sacd_dsdiff_t : public sacd_reader_t {
	abort_callback_impl m_abort;
public:
	service_ptr_t<file> m_file;
	uint32_t            m_version;
	uint32_t            m_samplerate;
	uint16_t            m_channel_count;
	int                 m_loudspeaker_config;
	int                 m_dst_encoded;
	uint64_t            m_frm8_size;
	uint64_t            m_dsti_offset;
	uint64_t            m_dsti_size;
	uint64_t            m_data_offset;
	uint64_t            m_data_size;
	uint16_t            m_framerate;
	uint32_t            m_frame_size;
	uint32_t            m_frame_count;
	array_t<subsong_t>  m_subsong;
	uint64_t            m_id3_offset;
	array_t<id3tags_t>  m_id3tags;
	bool                m_id3tags_indexed;
	uint32_t            m_current_subsong;                
	uint64_t            m_current_offset;
	uint64_t            m_current_size;
public:
	sacd_dsdiff_t(void);
	~sacd_dsdiff_t(void);
	uint32_t get_subsong_count();
	bool set_subsong(uint32_t subsong);
	int get_channels();
	int get_loudspeaker_config();
	int get_samplerate();
	uint64_t get_size();
	uint64_t get_offset();
	double get_duration();
	double get_duration(uint32_t subsong);
	bool is_dst();
	bool open(service_ptr_t<file> file, uint32_t mode = 0);
	bool close();
	bool read_frame(uint8_t* frame_data, size_t* frame_size);
	bool seek(double seconds);
	void get_info(uint32_t subsong, file_info& info);
	void set_info(uint32_t subsong, const file_info& info);
	bool commit();
private:
	uint64_t get_dsti_for_frame(uint32_t frame_nr);
	void get_id3tags(uint32_t subsong, file_info& info);
	void set_id3tags(uint32_t subsong, const file_info& info);
	void index_id3tags(file_info& info);
	void write_id3tag(const void* data, uint32_t size);
};

#endif
